#include <signal.h>
#include "libgraph.h"
#include <alloc.h>
#include <X11/keysym.h>


typedef struct
  {  
     int event_id;
     Window win;
     int x;
     int y;
     int button;
     int key;
  } EventType;


static EventType gr_queue_events[SIZE_QUEUE];
static int gr_head_events = 0;       /* position of next read */
static int gr_tail_events = 0;       /* position of next write */

#define EventQueueIsEmpty (gr_head_events == gr_tail_events)
#define EventQueueIsFull  (gr_head_events == gr_tail_events + 1)



void gr_enqueue_events(win,id_event,x,y,button,key)
     int id_event;
     Window win;
     int x,y,button,key;
{
  if (EventQueueIsFull) return;
  gr_queue_events[gr_tail_events].event_id = id_event;
  gr_queue_events[gr_tail_events].win=win;
  gr_queue_events[gr_tail_events].x=x;
  gr_queue_events[gr_tail_events].y=y;
  gr_queue_events[gr_tail_events].button=button;
  gr_queue_events[gr_tail_events].key=key;
  gr_tail_events++;
  if (gr_tail_events >= SIZE_QUEUE) gr_tail_events = 0;
}


int convert_key_code(c)
    int c;
{
  switch(c)
  {
    case XK_F1: return F1_code;
    case XK_F2: return F2_code;
    case XK_F3: return F3_code;
    case XK_F4: return F4_code;
    case XK_F5: return F5_code;
    case XK_F6: return F6_code;
    case XK_F7: return F7_code;
    case XK_F8: return F8_code;
    case XK_F9: return F9_code;
    case XK_F10: return F10_code;

    case XK_Right: return Right_code;
    case XK_Left: return Left_code;
    case XK_Up: return Up_code;
    case XK_Down: return Down_code;

    case XK_Next: return PageDown_code;
    case XK_Prior: return PageUp_code;
    case XK_Home: return Home_code;
    case XK_End: return End_code;

    case XK_Delete: return Del_code;
    case XK_Insert: return Ins_code;

    case XK_L6: return Copy_code;
    case XK_L8: return Past_code;
    case XK_L10: return Cut_code;

    default: return c;
  }
}


value gr_get_event(eventlist)
     value eventlist;
{
  value res;
  int mask;
  Bool poll,get_motion;
  int mouse_x, mouse_y, button, key,event_id, nu_win;
  Window rootwin, childwin;
  int root_x, root_y, win_x, win_y, modifiers;
  sighandler_return_type (*oldsigio)();
  XEvent event;
  int i,j;
  canvas *tmp;

  #ifdef DEBUG_EVENT
     fprintf(stderr,"gr_get_event :  begin\n");
  #endif


  mask = 0;
  poll = False;
  get_motion = False;
  while (Tag_val(eventlist) == 1) 
  {
    switch (Tag_val(Field(eventlist, 0))) 
    {
    case 0:                     /* Button_down */
      mask |= ButtonPressMask;
      mask |= PointerMotionMask; break;
    case 1:                     /* Button_up */
      mask |= ButtonReleaseMask; break;
    case 2:                     /* Key_pressed */
      mask |= KeyPressMask; break;
    case 3:                     /* Mouse_motion */
      mask |= PointerMotionMask; 
      get_motion = True; break;
    case 4:                     /* Poll */
      poll = True; break;
    case 5:                     /* Redraw */
    case 6:                     /* Close */
    case 7:                     /* Size */
      break;
    case 8:                     /* Time */
      signal(SIGALRM,gr_time_handler);
      alarm(gr_alarm_time);
      break;     
    }
    eventlist = Field(eventlist, 1);
  } 
  mouse_x = -1;
  mouse_y = -1;
  button = 0;
  key = 0x100;
  begin:


  if (!EventQueueIsEmpty) 
  {
    #ifdef DEBUG_EVENT
       fprintf(stderr,"EventQueue not empty\n");
    #endif

       nu_win = gr_queue_events[gr_head_events].win;
       mouse_x = gr_queue_events[gr_head_events].x;
       mouse_y = gr_queue_events[gr_head_events].y;
       button = gr_queue_events[gr_head_events].button;
       key = gr_queue_events[gr_head_events].key;
       event_id=gr_queue_events[gr_head_events].event_id;
       gr_head_events++;
       if (gr_head_events >= SIZE_QUEUE) gr_head_events = 0;
       if(event_id==Button_up && (mask & ButtonReleaseMask)==0) goto begin;
       if(event_id==Button_down && (mask & ButtonPressMask)==0) goto begin;
       if(event_id==Keypressed && (mask & KeyPressMask)==0) goto begin;
  } 
  else

  if (poll) 
  {
    if (XQueryPointer(grdisplay, grwindows[current_window].win,
                      &rootwin, &childwin,
                      &root_x, &root_y, &win_x, &win_y,
                      &modifiers)) 
    {
      mouse_x = win_x;
      mouse_y = win_y;
      event_id=Poll;
      nu_win=grwindows[current_window].win; 
      key=0;
    }
    button = modifiers & Button1Mask;
   } 
   else 
   {
        oldsigio = signal(SIGIO, SIG_IGN);

        again:

        XNextEvent(grdisplay, &event);
        switch(event.type) 
        {
          case Expose:
                 mouse_x = event.xexpose.x;
                 mouse_y = event.xexpose.y;
                 button = event.type == ButtonPress;
                 event_id=Redraw;
                 nu_win=event.xexpose.window;
                 break;

          case ButtonPress:
                 if((mask &  ButtonPressMask)==0)  goto again;
                 mouse_x = event.xbutton.x;
                 mouse_y = event.xbutton.y;
                 button = event.type == ButtonPress;
                 event_id=Button_down;
                 nu_win=event.xbutton.window;
                 break;

          case ButtonRelease:
                 if((mask &  ButtonReleaseMask)==0)  goto again;
                 mouse_x = event.xbutton.x;
                 mouse_y = event.xbutton.y;
                 button = event.type == ButtonPress;
                 event_id=Button_up;
                 nu_win=event.xbutton.window;
                 break;

          case MotionNotify:
                 if(!get_motion)  goto again;
                 mouse_x = event.xmotion.x;
                 mouse_y = event.xmotion.y;
                 button = event.xmotion.state & Button1Mask;
                 event_id=Motion;
                 nu_win=event.xmotion.window;
                 break;

         case KeyPress:
                 gr_handle_simple_event(&event);
                 if(EventQueueIsEmpty) goto again; /*Modifier key didn't fill the queue. */
                 if (!(mask & KeyPressMask)) goto again;
                 key = convert_key_code(gr_queue_events[gr_head_events].key);
                 nu_win = gr_queue_events[gr_head_events].win;
                 mouse_x = 0;
                 mouse_y = 0;
                 button = 0;
                 event_id=Keypressed;
                 gr_head_events++;
                 if (gr_head_events >= SIZE_QUEUE) gr_head_events = 0;
                 break;

         case DestroyNotify:
                 mouse_x = 0;
                 mouse_y = 0;
                 button = 0;
                 event_id=Close;
                 key=0;
                 nu_win=event.xdestroywindow.window;
                 break;

	 case ClientMessage:
                 mouse_x = 0;
                 mouse_y = 0;
                 button = 0;
                 event_id=Time;
                 key=0;
                 nu_win=event.xclient.window;
                 break;

         default:
                 gr_handle_simple_event(&event);
                 goto again;
      } /* end switch */

      signal(SIGIO, oldsigio);

      XFlush(grdisplay);
    }  /* end else */
 
  if(nb_windows == 0) goto begin;
  if((key == 0) && (event_id == Keypressed)) goto begin;

  for (i=0; i < nb_windows; i++)
    if(grwindows[i].win == nu_win) break;

  if(i != nb_windows) 
  {
    current_window = i;
    ValidWindow = TRUE;
    XSetInputFocus(grdisplay,grwindows[i].win,RevertToParent,CurrentTime);
  }
  else if(event_id == Keypressed) goto begin; /* ignore the bad keyboard focus*/

  alarm(0);

  #ifdef DEBUG_EVENT
     fprintf(stderr,"gr_get_event : win_id=%d, event_id=%d, mouse_x=%d, mouse_y=%d, button=%d, key=%d\n", nu_win, event_id, mouse_x, mouse_y, button, key);
  #endif


  res = alloc_tuple(6);
  Field(res, 0) = Val_int(nu_win);
  Field(res, 1) = Atom(event_id);
  Field(res, 2) = Val_int(mouse_x);
  Field(res, 3) = Val_int(mouse_y == -1 ? -1 : Wcvt(mouse_y));
  Field(res, 4) = Val_bool(button);
  Field(res, 5) = Val_int(key);
  return res;
}

value gr_send_event(Event)      /* ML */
  value Event;
{
  return Val_unit;
}


